let version = "dev"
let src = ""
let port = 80
let bindir = ""
let libdir = ""
let libdir_set = 0
let logdir = ""
let configdir = ""
let staticpagesdir = ""
let datadir = ""
let uploaddir = ""
let commandpipe = ""
let mimefile = ""
